/*
 * This file is part of the libopencm3 project.
 *
 * Copyright (C) 2009 Uwe Hermann <uwe@hermann-uwe.de>
 *
 * This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <errno.h>
#include <stdio.h>
#include <unistd.h>


#include <libopencm3/stm32/rcc.h>
#include <libopencm3/stm32/gpio.h>
#include <libopencm3/stm32/usart.h>
#include <libopencm3/cm3/systick.h>
#include <libopencm3/cm3/nvic.h>
#include <libopencm3/cm3/scb.h>

//port operations
#define PIN_SET(port, pins)	port->ODR |= (pins)			//set pins on port
#define PIN_CLR(port, pins)	port->ODR &=~(pins)			//clear pins on port
#define PIN_FLP(port, pins)	port->ODR ^= (pins)			//flip pins on port
#define PIN_GET(port, pins) (port->IDR & (pins))		//read pins on port

#define LED_PIN     GPIO4

typedef struct {
     char * buf;
     int head;
     int tail;
     int size;
} fifo_t;
 

#ifdef __cplusplus
extern "C" {
#endif

    //void sys_tick_handler(void);
    int _write(int file, const char *ptr, ssize_t len);
    void USART1_isr(void);
    static void systick_setup(void);
    void delay(uint64_t duration);
    uint64_t millis(void);
    void fifo_init(fifo_t * f, char * buf, int size);
    int fifo_read(fifo_t * f, void * buf, int nbytes);
    int fifo_write(fifo_t * f, const void * buf, int nbytes);

#ifdef __cplusplus
}
#endif





//**************************************************************************************
 
/*
// FIFO implementation from  http://coactionos.com/embedded%20design%20tips/2013/10/02/Tips-A-FIFO-Buffer-Implementation/
// Ivan's Thread  https://my.st.com/public/STe2ecommunities/mcu/Lists/STM32Discovery/Flat.aspx?RootFolder=%2fpublic%2fSTe2ecommunities%2fmcu%2fLists%2fSTM32Discovery%2fUART%20example%20code%20for%20STM32F0&FolderCTID=0x01200200770978C69A1141439FE559EB459D75800084C20D8867EAD444A5987D47BE638E0F&TopicsView=https%3A%2F%2Fmy.st.com%2Fpublic%2FSTe2ecommunities%2Fmcu%2FLists%2FSTM32Discovery%2FAllItems.aspx¤tviews=10407
*/



//This initializes the FIFO structure with the given buffer and size
void fifo_init(fifo_t * f, char * buf, int size){
     f->head = 0;
     f->tail = 0;
     f->size = size;
     f->buf = buf;
}
 
//This reads nbytes bytes from the FIFO
//The number of bytes read is returned
int fifo_read(fifo_t * f, void * buf, int nbytes){
     int i;
     char * p;
     p = buf;
     for(i=0; i < nbytes; i++){
          if( f->tail != f->head ){ //see if any data is available
               *p++ = f->buf[f->tail];  //grab a byte from the buffer
               f->tail++;  //increment the tail
               if( f->tail == f->size ){  //check for wrap-around
                    f->tail = 0;
               }
          } else {
               return i; //number of bytes read
          }
     }
     return nbytes;
}
 
//This writes up to nbytes bytes to the FIFO
//If the head runs in to the tail, not all bytes are written
//The number of bytes written is returned
int fifo_write(fifo_t * f, const void * buf, int nbytes){
     int i;
     const char * p;
     p = buf;
     for(i=0; i < nbytes; i++){
           //first check to see if there is space in the buffer
           if( (f->head + 1 == f->tail) ||
                ( (f->head + 1 == f->size) && (f->tail == 0) )){
                 return i; //no more room
           } else {
               f->buf[f->head] = *p++;
               f->head++;  //increment the head
               if( f->head == f->size ){  //check for wrap-around
                    f->head = 0;
               }
           }
     }
     return nbytes;
}
 
//**************************************************************************************
 
#define RXSIZE 1024
 
uint8_t RxBuffer[RXSIZE];
fifo_t RxFifo[1];
 
#define TXSIZE 512
 
uint8_t TxBuffer[TXSIZE];
fifo_t TxFifo[1];
 
//**************************************************************************************

static void clock_setup(void)
{
	rcc_clock_setup_in_hse_8mhz_out_48mhz();
    // First, let's ensure that our clock is running off the high-speed internal
    // oscillator (HSI) at 48MHz.
    //rcc_clock_setup_in_hsi_out_48mhz();

	/* Enable GPIOA clock. */
	rcc_periph_clock_enable(RCC_GPIOA);

    rcc_periph_clock_enable(RCC_GPIOB);

	/* Enable clocks for GPIO port A (for LED GPIO_USART1_TX) and USART1. */
	rcc_periph_clock_enable(RCC_USART1);
}

static void usart_setup(void)
{
#if 0
	/* Enable the USART1 interrupt. */
	nvic_enable_irq(NVIC_USART1_IRQ);

	/* Setup GPIO pin GPIO_USART1_RE_TX on GPIO port A for transmit. */
	gpio_set_mode(GPIOA, GPIO_MODE_OUTPUT_50_MHZ,
		      GPIO_CNF_OUTPUT_ALTFN_PUSHPULL, GPIO_USART1_TX);

	/* Setup GPIO pin GPIO_USART1_RE_RX on GPIO port A for receive. */
	gpio_set_mode(GPIOA, GPIO_MODE_INPUT,
		      GPIO_CNF_INPUT_FLOAT, GPIO_USART1_RX);

	/* Setup UART parameters. */
	usart_set_baudrate(USART1, 115200);
	usart_set_databits(USART1, 8);
	usart_set_stopbits(USART1, USART_STOPBITS_1);
	usart_set_parity(USART1, USART_PARITY_NONE);
	usart_set_flow_control(USART1, USART_FLOWCONTROL_NONE);
	usart_set_mode(USART1, USART_MODE_TX_RX);

	/* Enable USART1 Receive interrupt. */
	USART_CR1(USART1) |= USART_CR1_RXNEIE;

	/* Finally enable the USART. */
	usart_enable(USART1);
#endif


  // Initialize early
  fifo_init(RxFifo, (void *)RxBuffer, sizeof(RxBuffer));
  fifo_init(TxFifo, (void *)TxBuffer, sizeof(TxBuffer));

  // For the peripheral to work, we need to enable it's clock
  rcc_periph_clock_enable(RCC_USART1);
  // From the datasheet for the STM32F0 series of chips (Page 30, Table 11)
  // we know that the USART1 peripheral has it's TX line connected as
  // alternate function 1 on port A pin 9.
  // In order to use this pin for the alternate function, we need to set the
  // mode to GPIO_MODE_AF (alternate function). We also do not need a pullup
  // or pulldown resistor on this pin, since the peripheral will handle
  // keeping the line high when nothing is being transmitted.
  gpio_mode_setup(GPIOA, GPIO_MODE_AF, GPIO_PUPD_NONE, GPIO9 | GPIO10);
  // Now that we have put the pin into alternate function mode, we need to
  // select which alternate function to use. PA9 can be used for several
  // alternate functions - Timer 15, USART1 TX, Timer 1, and on some devices
  // I2C. Here, we want alternate function 1 (USART1_TX)
  gpio_set_af(GPIOA, GPIO_AF1, GPIO9 | GPIO10);
  // Now that the pins are configured, we can configure the USART itself.
  // First, let's set the baud rate at 115200
  usart_set_baudrate(USART1, 115200);
  // Each datum is 8 bits
  usart_set_databits(USART1, 8);
  // No parity bit
  usart_set_parity(USART1, USART_PARITY_NONE);
  // One stop bit
  usart_set_stopbits(USART1, USART_CR2_STOPBITS_1);
  // For a debug console, we only need unidirectional transmit
  usart_set_mode(USART1, USART_MODE_TX_RX);
  // No flow control
  usart_set_flow_control(USART1, USART_FLOWCONTROL_NONE);
  // Enable the peripheral
  usart_enable(USART1);

  // Enable interrupts for tx/rx
  nvic_enable_irq(NVIC_USART1_IRQ);
  usart_enable_rx_interrupt(USART1);
  usart_enable_tx_interrupt(USART1);
}

static void gpio_setup(void)
{
    // Our test LED is connected to Port A pin 04, so let's set it as output
    gpio_mode_setup(GPIOA, GPIO_MODE_OUTPUT, GPIO_PUPD_NONE, LED_PIN);
}


static void systick_setup(void) {
    // Set the systick clock source to our main clock
    systick_set_clocksource(STK_CSR_CLKSOURCE_AHB);
    // Clear the Current Value Register so that we start at 0
    STK_CVR = 0;
    // In order to trigger an interrupt every millisecond, we can set the reload
    // value to be the speed of the processor / 1000 -1
    systick_set_reload(rcc_ahb_frequency / 1000 - 1);
    // Enable interrupts from the system tick clock
    systick_interrupt_enable();
    // Enable the system tick counter
    systick_counter_enable();
}

// Storage for our monotonic system clock.
// Note that it needs to be volatile since we're modifying it from an interrupt.
static volatile uint64_t _millis = 0;

uint64_t millis(void) {
    return _millis;
}

// This is our interrupt handler for the systick reload interrupt.
// The full list of interrupt services routines that can be implemented is
// listed in libopencm3/include/libopencm3/stm32/f0/nvic.h
void sys_tick_handler(void) {
    // Increment our monotonic clock
    _millis++;
}

/**
 * Delay for a real number of milliseconds
 */
void delay(uint64_t duration) {
    const uint64_t until = millis() + duration;
    while (millis() < until);
}

#if 0
void USART1_isr(void)
{
	static uint8_t data = 'A';

	/* Check if we were called because of RXNE. */
	if (((USART_CR1(USART1) & USART_CR1_RXNEIE) != 0) &&
	    ((USART_SR(USART1) & USART_SR_RXNE) != 0)) {

		/* Indicate that we got data. */
		gpio_toggle(GPIOA, GPIO5);

		/* Retrieve the data from the peripheral. */
		data = usart_recv(USART1);

		/* Enable transmit interrupt so it sends back the data. */
		USART_CR1(USART1) |= USART_CR1_TXEIE;
	}

	/* Check if we were called because of TXE. */
	if (((USART_CR1(USART1) & USART_CR1_TXEIE) != 0) &&
	    ((USART_SR(USART1) & USART_SR_TXE) != 0)) {

		/* Indicate that we are sending out data. */
		// gpio_toggle(GPIOA, GPIO5);

		/* Put data into the transmit register. */
		usart_send(USART1, data);

		/* Disable the TXE interrupt as we don't need it anymore. */
		USART_CR1(USART1) &= ~USART_CR1_TXEIE;
	}
}
#endif

//void usart1_isr(void) { USART1_isr(); }

//void USART1_isr(void) {
void usart1_isr(void) {

    //gpio_toggle(GPIOA, LED_PIN);
  // Check if interrupt is signalling TX ready
  if (usart_get_flag(USART1, USART_FLAG_TXE)) {
    // Try and pop a byte from the tx buffer
    uint16_t tx;
    if (fifo_read(TxFifo, &tx, 1) == 1) // Anything to send
     {
      // Move the next byte of data to the tx buffer
      usart_send(USART1, tx);
    } else {
      // If there's no more data to be sent, disable this interrupt
      usart_disable_tx_interrupt(USART1);
    }
  }

  // Check if data has been received (RX Not Empty)
  if (usart_get_flag(USART1, USART_FLAG_RXNE)) {
    uint16_t rx = usart_recv(USART1);
    //const uint16_t c = usart_recv(USART1);
    // Might fail if buffer is full!
    fifo_write(RxFifo, &rx, 1); // Place in reception fifo
  }
}



int main(void)
{
	//SCB_VTOR = (uint32_t) 0x08005000;

	clock_setup();
	gpio_setup();
    systick_setup();
	usart_setup();
    
    uint64_t now,next=0;
    
    now = millis();
    next = now;
    gpio_set(GPIOA, LED_PIN);

    // Toggle the LED on and off forever
    while (1) 
    {

        uint8_t fwd;
 
        if (fifo_read(RxFifo, &fwd, 1) == 1) // Anything available in the reception buffer?
        {
          fifo_write(TxFifo, &fwd, 1); // Echo it back to be sent
          usart_enable_tx_interrupt(USART1); // Enable when something to send
          //gpio_toggle(GPIOA, LED_PIN);
        }
     
        now = millis();
        if((now - 500) >= next)
        {
            next = now;
            //printf("[%lld] LED on\n", millis());
            gpio_toggle(GPIOA, LED_PIN);
        }

    } //while(1)

	return 0;
}
